// DiskMod

#include <ntddk.h>
#include <ntddstor.h>
#include <ntdddisk.h>
#include <ntstrsafe.h>
#include <scsi.h>

#if DBG
 #define DBGPrint0(xx) DbgPrint xx
 #define DBGPrint1(xx) DbgPrint xx
 #define DBGPrint2(xx) DbgPrint xx
 #define DBGPrint3(xx) 0
 #define DBGPrint4(xx) 0
 #define DBGCODE(xx)   xx
 #define DBGCODE4(xx)  0
#else
 #define DBGPrint0(xx) DbgPrint xx
 #define DBGPrint1(xx) 0
 #define DBGPrint2(xx) 0
 #define DBGPrint3(xx) 0
 #define DBGPrint4(xx) 0
 #define DBGCODE(xx)   0
 #define DBGCODE4(xx)  0
#endif

enum {
    TagDiskMod = 'doMD',
};

typedef struct _DISKMOD_DEV_EXT {
	PDEVICE_OBJECT devobj;
	PDEVICE_OBJECT lowerDO;
	KEVENT pagingcount_lock;
	LONG   pagingcount_lower;
	LONG   pagingcount_override;
	ULONG  mbrsignature;
	UCHAR  configpaging;     // 0=disable 1=enable 2=unmodified
	UCHAR  configremovable;  // 0=fixed 1=removable 2=unmodified
} DISKMOD_DEV_EXT, *PDISKMOD_DEV_EXT;

typedef struct _WORK_CONTEXT {
	PIO_WORKITEM pwi;
	PIRP Irp;
} WORK_CONTEXT, *PWORK_CONTEXT;


// function declarations
NTSTATUS DriverEntry( IN PDRIVER_OBJECT  DriverObject, IN PUNICODE_STRING  RegistryPath );
DRIVER_UNLOAD     DiskModUnload; 
DRIVER_ADD_DEVICE DiskModAddDevice;

DRIVER_DISPATCH   DiskModDispatchSkip;
DRIVER_DISPATCH   DiskModDispatchPnP;
DRIVER_DISPATCH   DiskModDispatchPower;
DRIVER_DISPATCH   DiskModDispatchDeviceControl;
DRIVER_DISPATCH   DiskModDispatchSystemControl;
DRIVER_DISPATCH   DiskModDispatchCreateClose;
DRIVER_DISPATCH   DiskModDispatchReadWrite;
DRIVER_DISPATCH   DiskModDispatchSCSI;

IO_COMPLETION_ROUTINE DiskModIoCompletionPnPDeviceUsageNotification;

// ----------------------------------------
// global variables

UNICODE_STRING usDriverRegPath;

// ----------------------------------------
// non-pagable functions

NTSTATUS DiskModDispatchSkip( IN PDEVICE_OBJECT  DeviceObject, IN PIRP  Irp )
{
	PDISKMOD_DEV_EXT fdoext = (PDISKMOD_DEV_EXT)DeviceObject->DeviceExtension;
	PDEVICE_OBJECT ldo = fdoext->lowerDO;
	DBGCODE(
	PIO_STACK_LOCATION pisl = IoGetCurrentIrpStackLocation(Irp);
	DBGPrint4(("DiskMod: DispatchSkip %p %p %08X %08X\n", DeviceObject, Irp, pisl->MajorFunction, pisl->MinorFunction));
	);
	IoSkipCurrentIrpStackLocation(Irp);
	return IoCallDriver(ldo, Irp);
}

NTSTATUS DiskModDispatchPower( IN PDEVICE_OBJECT  DeviceObject, IN PIRP  Irp )
{
	PDISKMOD_DEV_EXT fdoext = (PDISKMOD_DEV_EXT)DeviceObject->DeviceExtension;
	PIO_STACK_LOCATION iost;
	NTSTATUS status;
	iost = IoGetCurrentIrpStackLocation(Irp);
	switch (iost->MinorFunction)
	{
	case IRP_MN_QUERY_POWER:	//Optional, top-down
	case IRP_MN_SET_POWER:
		PoStartNextPowerIrp(Irp); // must call this for each IRP_MN_QUERY_POWER, IRP_MN_SET_POWER in Win2K-2003
		// then pass IRP down
		break;
	}
	// default -- pass IRP down the I/O stack
	IoSkipCurrentIrpStackLocation(Irp);
	status = PoCallDriver(fdoext->lowerDO, Irp);
	return status;
}

NTSTATUS DiskModDispatchSCSI( IN PDEVICE_OBJECT  DeviceObject, IN PIRP  Irp )
{
	PDISKMOD_DEV_EXT fdoext = (PDISKMOD_DEV_EXT)DeviceObject->DeviceExtension;
	PDEVICE_OBJECT ldo = fdoext->lowerDO;
	DBGCODE4(
		PIO_STACK_LOCATION pisl = IoGetCurrentIrpStackLocation(Irp);
		PSCSI_REQUEST_BLOCK psrb = pisl->Parameters.Scsi.Srb;
		DBGPrint4(("DiskMod: DispatchSCSI %p %p %08X %08X  %p %02X %02X\n", DeviceObject, Irp, pisl->MajorFunction, pisl->MinorFunction,
			psrb, psrb->Function, psrb->Cdb[0] ));
	);
	IoSkipCurrentIrpStackLocation(Irp);
	return IoCallDriver(ldo, Irp);
}

NTSTATUS CallDriverSyncIoCompletion( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context )
{
	PKEVENT pev = (PKEVENT) Context;
	DBGPrint1(("DiskMod: CallDriverSyncIoCompletion %p %p %p\n", DeviceObject, Irp, Context));
	KeSetEvent(pev, 0, FALSE);
	return STATUS_MORE_PROCESSING_REQUIRED;
}

// ----------------------------------------
// pagable functions

void HexFromUlongW(PWCHAR pws, ULONG number)
{
	ULONG n = number;
	UCHAR i;
	for (i=8; i>0; n>>=4)
		pws[--i] = L"0123456789ABCDEF" [n&15];	
}

NTSTATUS CallDriverSync( IN PDEVICE_OBJECT  TargetDeviceObject, IN PIRP  Irp )
{
	KEVENT ev;
	NTSTATUS status;
	DBGCODE(
	PIO_STACK_LOCATION pisl = IoGetNextIrpStackLocation(Irp);
	DBGPrint4(("DiskMod: CallDriverSync %p %p %08X %08X\n", TargetDeviceObject, Irp,
		pisl->MajorFunction, pisl->MinorFunction));
	);
	KeInitializeEvent(&ev, NotificationEvent, FALSE);
	IoSetCompletionRoutine(Irp, CallDriverSyncIoCompletion, &ev, TRUE,TRUE,TRUE);
	status = IoCallDriver(TargetDeviceObject, Irp);
	if (status == STATUS_PENDING)
	{
		DBGPrint4(("DiskMod: CallDriverSync waiting\n",status));
		KeWaitForSingleObject(&ev, Executive, KernelMode, FALSE, NULL);
		status = Irp->IoStatus.Status;
	}
	DBGPrint1(("DiskMod: CallDriverSync end status=%08X\n",status));
	return status;
}

NTSTATUS ForwardIrpSync( IN PDEVICE_OBJECT  DeviceObject, IN PIRP  Irp )
{
	PDISKMOD_DEV_EXT fdoext = (PDISKMOD_DEV_EXT)DeviceObject->DeviceExtension;
	DBGCODE(
	PIO_STACK_LOCATION pisl = IoGetCurrentIrpStackLocation(Irp);
	DBGPrint4(("DiskMod: ForwardIrpSync %p %p %08X %08X\n",DeviceObject, Irp, 
		pisl->MajorFunction, pisl->MinorFunction));
	);
	IoCopyCurrentIrpStackLocationToNext(Irp);
	return CallDriverSync(fdoext->lowerDO, Irp);
}

typedef struct _READ_DISK_SIGNATURE_CONTEXT
{
	CHAR buffer[2048];
	SCSI_REQUEST_BLOCK srb;
	IRP irp; // must be at end of structure, followed by IO_STACK_LOCATION
} READ_DISK_SIGNATURE_CONTEXT, *PREAD_DISK_SIGNATURE_CONTEXT;

NTSTATUS DiskModReadDiskSignature( IN PDEVICE_OBJECT DeviceObject, PULONG signature )
{
	PDISKMOD_DEV_EXT fdoext = (PDISKMOD_DEV_EXT)DeviceObject->DeviceExtension;
	NTSTATUS status = STATUS_SUCCESS;
	CCHAR stacksize = fdoext->lowerDO->StackSize;
	SIZE_T size = FIELD_OFFSET(READ_DISK_SIGNATURE_CONTEXT,irp)+IoSizeOfIrp(stacksize);
	PREAD_DISK_SIGNATURE_CONTEXT pcontext = ExAllocatePoolWithTag(NonPagedPool, size, TagDiskMod);
	if (pcontext)
	{
		PIRP pirp;
		PMDL pmdl;
		PIO_STACK_LOCATION piosln;
		PSCSI_REQUEST_BLOCK psrb;
		PCDB pcdb;
		RtlZeroMemory(pcontext, size);
		pirp = &pcontext->irp;
		IoInitializeIrp(pirp, sizeof(pcontext->irp), stacksize);
		pmdl = IoAllocateMdl(pcontext->buffer, 512, FALSE, FALSE, pirp);
		if (pmdl)
		{
			MmBuildMdlForNonPagedPool(pmdl);
			pirp->MdlAddress = pmdl;
			pirp->AssociatedIrp.SystemBuffer = pcontext->buffer;

			piosln = IoGetNextIrpStackLocation(pirp);
			piosln->MajorFunction = IRP_MJ_SCSI;
			piosln->Parameters.Scsi.Srb = psrb = &pcontext->srb;

			psrb->Length = sizeof(pcontext->srb);
			psrb->Function = SRB_FUNCTION_EXECUTE_SCSI;
			psrb->DataTransferLength = 512;
			psrb->DataBuffer = pcontext->buffer;
			psrb->SrbFlags = SRB_FLAGS_DATA_IN|SRB_FLAGS_ADAPTER_CACHE_ENABLE;
			psrb->OriginalRequest = pirp;
			psrb->CdbLength = 10;
			pcdb = (PCDB) psrb->Cdb;
			pcdb->CDB10.OperationCode = SCSIOP_READ;
			pcdb->CDB10.LogicalBlockByte0 = 0;
			pcdb->CDB10.LogicalBlockByte1 = 0;
			pcdb->CDB10.LogicalBlockByte2 = 0;
			pcdb->CDB10.LogicalBlockByte3 = 0;
			pcdb->CDB10.TransferBlocksMsb = 0;
			pcdb->CDB10.TransferBlocksLsb = 1;
			// call lower driver
			status = CallDriverSync(fdoext->lowerDO, pirp);
			// result
			if (NT_SUCCESS(status))
			{
				DBGPrint4(("DiskMod: [0x1B8]=0x%08X  [0x1FE]=0x%04X\n",*(PUINT32)(pcontext->buffer+0x1B8),*(PUINT16)(pcontext->buffer+0x1FE) ));
				*signature = *(PUINT32)(pcontext->buffer+0x1B8);
			}
			else
			{
				DBGPrint1(("DiskMod: read failed status %08X %02X %02X\n",status,psrb->SrbStatus,psrb->ScsiStatus));
			}
			// clean up
			IoFreeMdl(pmdl);
		} else {
			DBGPrint1(("DiskMod: ReadDiskSignature allocate mdl failed\n"));
			status = STATUS_INSUFFICIENT_RESOURCES;
		}
		ExFreePoolWithTag(pcontext, TagDiskMod);		
		return status;
	} else {
		DBGPrint1(("DiskMod: ReadDiskSignature allocate memory failed\n"));
		return STATUS_INSUFFICIENT_RESOURCES;
	}
}

void DiskModReadConfig( IN PDEVICE_OBJECT  DeviceObject, IN HANDLE hkParam, ULONG level )
{
	PDISKMOD_DEV_EXT fdoext = (PDISKMOD_DEV_EXT)DeviceObject->DeviceExtension;
	NTSTATUS status;
	UNICODE_STRING valuename;
	struct {
		KEY_VALUE_PARTIAL_INFORMATION kvpi;
		UCHAR padding[7];
	} kvi;
	ULONG length=0;
	RtlInitUnicodeString(&valuename, L"Paging");
	status = ZwQueryValueKey(hkParam, &valuename, KeyValuePartialInformation, &kvi, sizeof(kvi), &length);
	if (NT_SUCCESS(status)) {
		if (kvi.kvpi.Type==REG_DWORD && kvi.kvpi.DataLength>=4) {
			ULONG data = *(PULONG)kvi.kvpi.Data;
			fdoext->configpaging = (UCHAR)data;
		}
	}
	RtlInitUnicodeString(&valuename, L"Removable");
	status = ZwQueryValueKey(hkParam, &valuename, KeyValuePartialInformation, &kvi, sizeof(kvi), &length);
	if (NT_SUCCESS(status)) {
		if (kvi.kvpi.Type==REG_DWORD && kvi.kvpi.DataLength>=4) {
			ULONG data = *(PULONG)kvi.kvpi.Data;
			fdoext->configremovable = (UCHAR)data;
		}
	}
	RtlInitUnicodeString(&valuename, L"BusType");
	status = ZwQueryValueKey(hkParam, &valuename, KeyValuePartialInformation, &kvi, sizeof(kvi), &length);
	if (NT_SUCCESS(status)) {
		if (kvi.kvpi.Type==REG_DWORD && kvi.kvpi.DataLength>=4) {
			ULONG data = *(PULONG)kvi.kvpi.Data;
			fdoext->configremovable = (UCHAR)data;
		}
	}
	// read disk specific parameters
	if (level > 0)
	{
		HANDLE hkSub;
		UNICODE_STRING keyname;
		WCHAR name_buffer[16];
		OBJECT_ATTRIBUTES oa;
		// open subkey
		RtlInitEmptyUnicodeString(&keyname, name_buffer, sizeof(name_buffer));
		RtlUnicodeStringCopyString(&keyname, L"MBR_00000000");
		HexFromUlongW(name_buffer+4, fdoext->mbrsignature);
		InitializeObjectAttributes(&oa,&keyname,OBJ_CASE_INSENSITIVE|OBJ_KERNEL_HANDLE,hkParam,NULL);
		status = ZwOpenKey(&hkSub, KEY_READ, &oa);
		if (NT_SUCCESS(status))
		{
			DBGPrint1(("DiskMod: open subkey %wZ successful\n",&keyname));
			// read config
			DiskModReadConfig(DeviceObject, hkSub, level-1);
			ZwClose(hkSub);
		}
		else
		{
			DBGPrint1(("DiskMod: open subkey %wZ failed\n",&keyname));
		}
	}
}

void DiskModStart( IN PDEVICE_OBJECT  DeviceObject )
{
	PDISKMOD_DEV_EXT fdoext = (PDISKMOD_DEV_EXT)DeviceObject->DeviceExtension;
	NTSTATUS status;
	// read signature
	status = DiskModReadDiskSignature(DeviceObject, &fdoext->mbrsignature);
	if (NT_SUCCESS(status)) {
		DBGPrint1(("DiskMod: MBR signature %08X\n",fdoext->mbrsignature));
	} else {
		DBGPrint1(("DiskMod: Read signature failed %08X\n",status));
	}			
	// read driver config
	{
		SIZE_T size = usDriverRegPath.Length + 11*sizeof(WCHAR);
		PVOID pnamebuffer = ExAllocatePoolWithTag(PagedPool, size, TagDiskMod);
		if (pnamebuffer)
		{
			HANDLE hkParam;
			UNICODE_STRING keyname;
			OBJECT_ATTRIBUTES oa;
			RtlInitEmptyUnicodeString(&keyname, pnamebuffer, (USHORT)size);
			RtlCopyUnicodeString(&keyname, &usDriverRegPath);
			RtlUnicodeStringCatString(&keyname, L"\\Parameters");
			InitializeObjectAttributes(&oa,&keyname,OBJ_CASE_INSENSITIVE|OBJ_KERNEL_HANDLE,NULL,NULL);
			status = ZwOpenKey(&hkParam, KEY_READ, &oa);
			if (NT_SUCCESS(status)) {
				DBGPrint1(("DiskMod: open key %wZ success\n",&keyname));
				DiskModReadConfig(DeviceObject, hkParam, 1);
				ZwClose(hkParam);
			} else {
				DBGPrint1(("DiskMod: open key %wZ failed\n",&keyname));
			}
			ExFreePoolWithTag(pnamebuffer,TagDiskMod);
		}
	}
	// read device config
	{
		HANDLE hkDeviceParam;
		status = IoOpenDeviceRegistryKey(DeviceObject, PLUGPLAY_REGKEY_DEVICE, KEY_READ, &hkDeviceParam);
		if (NT_SUCCESS(status)) {
			HANDLE hkParam;
			UNICODE_STRING keyname;
			OBJECT_ATTRIBUTES oa;
			RtlInitUnicodeString(&keyname, L"DiskMod");
			InitializeObjectAttributes(&oa,&keyname,OBJ_CASE_INSENSITIVE|OBJ_KERNEL_HANDLE,hkDeviceParam,NULL);
			status = ZwOpenKey(&hkParam, KEY_READ, &oa);
			if (NT_SUCCESS(status)) {
				DBGPrint1(("DiskMod: open subkey %wZ success\n",&keyname));
				DiskModReadConfig(DeviceObject, hkParam, 1);
				ZwClose(hkParam);
			} else {
				DBGPrint1(("DiskMod: open subkey %wZ failed\n",&keyname));
			}
			ZwClose(hkDeviceParam);
		} else {
			DBGPrint1(("DiskMod: open device key failed\n"));
		}
	}
}

NTSTATUS DiskModDispatchPnP( IN PDEVICE_OBJECT  DeviceObject, IN PIRP  Irp )
{
	PDISKMOD_DEV_EXT fdoext = (PDISKMOD_DEV_EXT)DeviceObject->DeviceExtension;
	PIO_STACK_LOCATION iost;
	NTSTATUS status;
	iost = IoGetCurrentIrpStackLocation(Irp);
	status = Irp->IoStatus.Status;
	DBGPrint1(("DiskMod: DispatchPnP %p %p %08X\n", DeviceObject, Irp, iost->MinorFunction ));
	switch (iost->MinorFunction)
	{
	case IRP_MN_START_DEVICE:			//Required, bottom-up
		DBGPrint1(("DiskMod:  START_DEVICE\n"));
		status = ForwardIrpSync(DeviceObject, Irp);
		DBGPrint1(("DiskMod:  START_DEVICE lower driver returned %08X\n",status));
		if (NT_SUCCESS(status))
		{
			DiskModStart(DeviceObject);
		}
		IoCompleteRequest(Irp,IO_NO_INCREMENT);
		DBGPrint1(("DiskMod:  START_DEVICE end\n"));
		return status;
	case IRP_MN_QUERY_STOP_DEVICE:		//Required, top-down
		DBGPrint1(("DiskMod:  QUERY_STOP_DEVICE\n"));
		break;
		// Win2K+: Stop for rebalancing resource. 
	case IRP_MN_STOP_DEVICE:			//Required, top-down
		DBGPrint1(("DiskMod:  STOP_DEVICE\n"));
		break;
	case IRP_MN_CANCEL_STOP_DEVICE:		//Required, bottom-up
		DBGPrint1(("DiskMod:  CANCEL_STOP_DEVICE\n"));
		break;
	case IRP_MN_QUERY_REMOVE_DEVICE:	//Required, top-down
		DBGPrint1(("DiskMod:  QUERY_REMOVE_DEVICE\n"));
		break;
	case IRP_MN_SURPRISE_REMOVAL:		//Required, top-down
		DBGPrint1(("DiskMod:  SURPRISE_REMOVAL\n"));
		break;
	case IRP_MN_REMOVE_DEVICE:			//Required, top-down
		DBGPrint1(("DiskMod:  REMOVE_DEVICE\n"));
		// call lower driver
		status = ForwardIrpSync(DeviceObject, Irp);
		// delete device
		IoDetachDevice(fdoext->lowerDO);
		IoDeleteDevice(DeviceObject);
		IoCompleteRequest(Irp, IO_NO_INCREMENT);
		DBGPrint1(("DiskMod:  REMOVE_DEVICE end\n"));
		return status;
	case IRP_MN_CANCEL_REMOVE_DEVICE:	//Required, bottom-up
		DBGPrint1(("DiskMod:  CANCEL_REMOVE_DEVICE\n"));
		break;
	//case IRP_MN_QUERY_CAPABILITIES:			//Optional
	//case IRP_MN_QUERY_PNP_DEVICE_STATE:		//Optional
	//case IRP_MN_FILTER_RESOURCE_REQUIREMENTS:	//Optional
	//	break;
	case IRP_MN_DEVICE_USAGE_NOTIFICATION:	//Required
		DBGPrint1(("DiskMod:  DEVICE_USAGE_NOTIFICATION\n"));
		if (iost->Parameters.UsageNotification.Type == DeviceUsageTypePaging)
		{
			if (iost->Parameters.UsageNotification.InPath)
			{	// in path
				if (fdoext->configpaging > 1)
				{	// Not override. 
					status = ForwardIrpSync(DeviceObject, Irp);
					KeWaitForSingleObject(&fdoext->pagingcount_lock, Executive, KernelMode, FALSE, NULL);
					if (NT_SUCCESS(status))
						InterlockedIncrement(&fdoext->pagingcount_lower);
					KeSetEvent(&fdoext->pagingcount_lock, 0, FALSE);
					IoCompleteRequest(Irp, IO_NO_INCREMENT);
					return status;
				}
				else if (fdoext->configpaging == 1)
				{	// Paging override enabled. 
					status = ForwardIrpSync(DeviceObject, Irp);
					KeWaitForSingleObject(&fdoext->pagingcount_lock, Executive, KernelMode, FALSE, NULL);
					if (NT_SUCCESS(status))
						InterlockedIncrement(&fdoext->pagingcount_lower);
					else
						InterlockedIncrement(&fdoext->pagingcount_override);
					KeSetEvent(&fdoext->pagingcount_lock, 0, FALSE);
					if (!NT_SUCCESS(status)) {
						Irp->IoStatus.Status = status = STATUS_SUCCESS;
					}
					IoCompleteRequest(Irp, IO_NO_INCREMENT);
					return status;
				}
				else // fdoext->configpaging == 0
				{	// Paging disabled. Fail. 
					Irp->IoStatus.Status = status = STATUS_UNSUCCESSFUL;
					IoCompleteRequest(Irp, IO_NO_INCREMENT);
					return status;
				}
			} else 
			{ // not in path
				UCHAR bOverride;
				KeWaitForSingleObject(&fdoext->pagingcount_lock, Executive, KernelMode, FALSE, NULL);
				if (fdoext->pagingcount_override > 0) {
					InterlockedDecrement(&fdoext->pagingcount_override);
					bOverride = TRUE;
				} else {
					InterlockedDecrement(&fdoext->pagingcount_lower);
					bOverride = FALSE;
				}
				KeSetEvent(&fdoext->pagingcount_lock, 0, FALSE);
				if (bOverride) {
					Irp->IoStatus.Status = status = STATUS_SUCCESS;
					IoCompleteRequest(Irp, IO_NO_INCREMENT);
					return status;
				} else {
					// Pass IRP down to lower driver
					break;
				}
			}
		}
		// non-paging usage notification, just pass IRP down device stack
		break;
	//case IRP_MN_QUERY_DEVICE_RELATIONS:		//Required
	//	break;
	//case IRP_MN_QUERY_RESOURCES:				//No
	//case IRP_MN_QUERY_RESOURCE_REQUIREMENTS:	//No
	//case IRP_MN_QUERY_ID:						//No
	//case IRP_MN_QUERY_DEVICE_TEXT:			//No
	//case IRP_MN_QUERY_BUS_INFORMATION:		//No
	//case IRP_MN_QUERY_INTERFACE:			//Optional
	//case IRP_MN_READ_CONFIG:					//No
	//case IRP_MN_WRITE_CONFIG:					//No
	//case IRP_MN_EJECT:						//No
	//case IRP_MN_SET_LOCK:						//No
	//	break;
	}
	// default -- pass IRP down the I/O stack
	IoSkipCurrentIrpStackLocation(Irp);
	status = IoCallDriver(fdoext->lowerDO, Irp);
	return status;
}

NTSTATUS DiskModDispatchDeviceControl( IN PDEVICE_OBJECT  DeviceObject, IN PIRP  Irp )
{
	PDISKMOD_DEV_EXT fdoext = (PDISKMOD_DEV_EXT)DeviceObject->DeviceExtension;
	PIO_STACK_LOCATION iost;
	NTSTATUS status;
	iost = IoGetCurrentIrpStackLocation(Irp);
	status = Irp->IoStatus.Status;
	DBGPrint1(("DiskMod: DispatchDeviceControl %p %p %08X\n", DeviceObject, Irp,
		iost->Parameters.DeviceIoControl.IoControlCode));
	switch ( iost->Parameters.DeviceIoControl.IoControlCode )
	{
	case IOCTL_STORAGE_QUERY_PROPERTY:			//Required, bottom-up
		DBGPrint1(("DiskMod:  QUERY_PROPERTY\n"));
		if (iost->Parameters.DeviceIoControl.InputBufferLength >= sizeof(STORAGE_PROPERTY_QUERY)
			&& Irp->AssociatedIrp.SystemBuffer)
		{
			PSTORAGE_PROPERTY_QUERY pspq = (PSTORAGE_PROPERTY_QUERY) Irp->AssociatedIrp.SystemBuffer;
			DBGPrint1(("DiskMod:   %d %d\n",pspq->PropertyId,pspq->QueryType));
			if (pspq->PropertyId == StorageDeviceProperty && pspq->QueryType == PropertyStandardQuery)
			{
				status = ForwardIrpSync(DeviceObject, Irp);
				if (NT_SUCCESS(status)) 
				{
					PSTORAGE_DEVICE_DESCRIPTOR p = (PSTORAGE_DEVICE_DESCRIPTOR)	Irp->AssociatedIrp.SystemBuffer;
					if (iost->Parameters.DeviceIoControl.OutputBufferLength >= RTL_SIZEOF_THROUGH_FIELD(STORAGE_DEVICE_DESCRIPTOR,RemovableMedia) )
					{
						UCHAR old_removable = p->RemovableMedia;
						if (fdoext->configremovable < 2) 
							p->RemovableMedia = (fdoext->configremovable)? TRUE: FALSE;
						DBGPrint1(("DiskMod:   removablemedia %d %d\n",old_removable,p->RemovableMedia));
					}
				}
				IoCompleteRequest(Irp,IO_NO_INCREMENT);
				return status;
			}
		}
		break;
	}
	// default -- pass IRP down the I/O stack
	IoSkipCurrentIrpStackLocation(Irp);
	status = IoCallDriver(fdoext->lowerDO, Irp);
	return status;
}

NTSTATUS DiskModAddDevice( IN PDRIVER_OBJECT  DriverObject, IN PDEVICE_OBJECT  PhysicalDeviceObject )
{
	PDEVICE_OBJECT fdo,ldo;
	NTSTATUS status;
	DBGPrint1(("DiskMod: AddDevice start\n"));
	status = IoCreateDevice( DriverObject, sizeof(DISKMOD_DEV_EXT), NULL,
		PhysicalDeviceObject->DeviceType, 
		PhysicalDeviceObject->Characteristics,
		FALSE, &fdo);
	if (!NT_SUCCESS(status)) {
		DBGPrint1(("DiskMod: AddDevice create device failed status=0x%X\n",status));
		return status;
	} else {
		PDISKMOD_DEV_EXT fdoext = (PDISKMOD_DEV_EXT)(fdo->DeviceExtension);
		fdoext->devobj = fdo;
		fdoext->mbrsignature = 0;
		fdoext->configpaging = 0xFF;
		fdoext->configremovable = 0xFF;
		fdoext->pagingcount_lower = 0;
		fdoext->pagingcount_override = 0;
		KeInitializeEvent(&fdoext->pagingcount_lock, SynchronizationEvent, TRUE);
		fdoext->lowerDO = ldo = IoAttachDeviceToDeviceStack(fdo, PhysicalDeviceObject);
		if (ldo==NULL) {
			DBGPrint1(("DiskMod: AddDevice attach to device stack failed\n"));
			IoDeleteDevice(fdo);
			return STATUS_UNSUCCESSFUL;
		}
		fdo->Characteristics &= ~(FILE_REMOVABLE_MEDIA);
		fdo->AlignmentRequirement = ldo->AlignmentRequirement;
		fdo->Flags |= (ldo->Flags & (DO_BUFFERED_IO|DO_DIRECT_IO)) | DO_POWER_INRUSH;
		DBGPrint1(("DiskMod: Characteristic %08X %08X\n",fdo->Characteristics,ldo->Characteristics));
		DBGPrint1(("DiskMod: Flags          %08X %08X\n",fdo->Flags,ldo->Flags));
		DBGPrint1(("DiskMod: DeviceType     %08X %08X\n",fdo->DeviceType,ldo->DeviceType));
		DBGPrint1(("DiskMod: StackSize      %8d %8d\n",fdo->StackSize,ldo->StackSize));
		DBGPrint1(("DiskMod: SectorSize     %8d %8d\n",fdo->SectorSize,ldo->SectorSize));
		// initialize complete
		fdo->Flags &= ~DO_DEVICE_INITIALIZING;
		status = STATUS_SUCCESS;
		DBGPrint1(("DiskMod: AddDevice success PDO=%p LDO=%p FDO=%p\n",PhysicalDeviceObject,fdoext->lowerDO,fdo));
	}
	return status;
}

VOID DiskModUnload( IN PDRIVER_OBJECT  DriverObject )
{
	// free driver-wide allocated resource
	DBGPrint1(("DiskMod: Unload ----------\n\n"));
	if (usDriverRegPath.Buffer) {
		ExFreePoolWithTag(usDriverRegPath.Buffer,TagDiskMod);
		RtlInitEmptyUnicodeString(&usDriverRegPath, NULL, 0);
	}
}

// ----------------------------------------
// initialization functions

NTSTATUS DriverEntry( IN PDRIVER_OBJECT  DriverObject, IN PUNICODE_STRING  RegistryPath )
{
	DBGPrint1(("DiskMod: DriverEntry start ----------\n"));

	// initialize DriverObject
	DriverObject->DriverExtension->AddDevice = DiskModAddDevice;
	//DriverObject->DriverStartIo = DiskModStartIo; 
	DriverObject->DriverUnload  = DiskModUnload; 
	DriverObject->MajorFunction[IRP_MJ_CREATE] = DiskModDispatchSkip;
	DriverObject->MajorFunction[IRP_MJ_CLOSE ] = DiskModDispatchSkip;
	DriverObject->MajorFunction[IRP_MJ_READ  ] = DiskModDispatchSkip;
	DriverObject->MajorFunction[IRP_MJ_WRITE ] = DiskModDispatchSkip;
	DriverObject->MajorFunction[IRP_MJ_PNP   ] = DiskModDispatchPnP; 
	DriverObject->MajorFunction[IRP_MJ_POWER ] = DiskModDispatchPower; 
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DiskModDispatchDeviceControl;
	DriverObject->MajorFunction[IRP_MJ_SYSTEM_CONTROL] = DiskModDispatchSkip;
	DriverObject->MajorFunction[IRP_MJ_SCSI  ] = DiskModDispatchSCSI;

	// save registry path
	if (usDriverRegPath.Buffer == NULL)
	{
		PVOID p =(PWCH)ExAllocatePoolWithTag(NonPagedPool, RegistryPath->Length, TagDiskMod);
		if (p)
		{
			RtlInitEmptyUnicodeString(&usDriverRegPath, (PWSTR)p, RegistryPath->Length);
			RtlCopyUnicodeString(&usDriverRegPath, RegistryPath);
		}
	}

	DBGPrint1(("DiskMod: DriverEntry success\n"));
	return STATUS_SUCCESS; 
}
